import os
import platform
import shutil
import sys
from glob import glob
from subprocess import DEVNULL, run


def main():
    root_dir = os.path.dirname(os.path.realpath(__file__))
    src_dir = os.path.join(root_dir, "src")
    tmp_dir = os.path.join(root_dir, "build")
    res_dir = os.path.join(root_dir, "res")

    assert len(sys.argv) > 1

    if sys.argv[1] == "--all":
        fns = glob(os.path.join(src_dir, "**/*.tex"), recursive=True)
    else:
        fns = sys.argv[1:]

    paths = [os.path.realpath(os.path.abspath(fn)) for fn in fns]

    os.makedirs(tmp_dir, exist_ok=True)

    for path in paths:
        assert os.path.splitext(path)[1] == ".tex"
        assert os.path.isfile(path)
        assert os.path.dirname(path).startswith(src_dir)

        subpath = path[len(src_dir) + 1 :]
        tmp_path = os.path.join(tmp_dir, subpath)

        os.makedirs(os.path.dirname(tmp_path), exist_ok=True)

        shutil.copy(path, tmp_path)
        os.chdir(os.path.dirname(tmp_path))

        run(
            [
                "pdflatex",
                "-interaction=nonstopmode",
                "-halt-on-error",
                os.path.basename(path),
            ],
            stdout=DEVNULL,
            check=True,
        )

        pdf_path = os.path.splitext(tmp_path)[0] + ".pdf"
        out_ext = ".png"
        out_path = os.path.join(res_dir, os.path.splitext(subpath)[0] + out_ext)

        os.makedirs(os.path.dirname(out_path), exist_ok=True)

        cmd = "convert" if platform.system().lower() == "linux" else "magick"

        run(
            [
                cmd,
                "-density",
                "400",
                pdf_path,
                out_path,
            ],
            stdout=DEVNULL,
            check=True,
        )


if __name__ == "__main__":
    main()



from pathlib import Path

import attr
import numpy as np
import pytest

from acconeer.exptool import a111
from acconeer.exptool.a111._clients.mock.client import MockClient


@pytest.fixture
def mocker():
    mocker = MockClient()
    mocker.squeeze = False
    return mocker


@pytest.mark.parametrize("mode", a111.Mode)
@pytest.mark.parametrize("ext", ["h5", "npz"])
@pytest.mark.parametrize("give_pathlib_path", [True, False])
def test_recording(tmp_path, mode, ext, give_pathlib_path, mocker):
    config = a111._configs.MODE_TO_CONFIG_CLASS_MAP[mode]()
    config.downsampling_factor = 2

    session_info = mocker.start_session(config)

    recorder = a111.recording.Recorder(
        sensor_config=config,
        session_info=session_info,
    )

    for _ in range(10):
        data_info, data = mocker.get_next()
        recorder.sample(data_info, data)

    recorder.close()
    record = recorder.record

    assert record.mode == mode
    assert record.sensor_config_dump == config._dumps()
    assert len(record.data) == 10
    assert len(record.data_info) == 10
    assert isinstance(record.data, np.ndarray)

    filename = Path(tmp_path).joinpath("record." + ext)

    if not give_pathlib_path:
        filename = str(filename)

    a111.recording.save(filename, record)
    loaded_record = a111.recording.load(filename)

    for a in attr.fields(a111.recording.Record):
        assert np.all(getattr(record, a.name) == getattr(loaded_record, a.name))

    assert record.sensor_config.downsampling_factor == config.downsampling_factor


def test_unknown_mode(mocker):
    config = a111.EnvelopeServiceConfig()
    session_info = mocker.start_session(config)
    recorder = a111.recording.Recorder(sensor_config=config, session_info=session_info)
    data_info, data = mocker.get_next()
    recorder.sample(data_info, data)
    recorder.close()
    record = recorder.record

    packed = a111.recording.pack(record)
    assert "mode" in packed

    with pytest.warns(None) as captured_warnings:
        a111.recording.unpack(packed)

    assert len(captured_warnings) == 0

    with pytest.warns(UserWarning):
        packed["mode"] = "some_unknown_mode"
        unpacked = a111.recording.unpack(packed)

    assert unpacked.mode is None


@pytest.mark.parametrize("mode", a111.Mode)
def test_packing_mode(tmp_path, mode):
    config = a111._configs.MODE_TO_CONFIG_CLASS_MAP[mode]()
    config.downsampling_factor = 2

    sensor_config_dump = config._dumps()
    session_info = {"foo": "bar"}

    record = a111.recording.Record(
        mode=mode,
        sensor_config_dump=sensor_config_dump,
        session_info=session_info,
        data=[],
        data_info=[],
    )

    packed = a111.recording.pack(record)
    restored = a111.recording.unpack(packed)

    assert restored.mode == mode



project = "acconeer-python-exploration"
copyright = "2019 - 2022, Acconeer AB"
author = "Acconeer AB"

# version = ""  # The short X.Y version
# release = ""  # The full version, including alpha/beta/rc tags

extensions = [
    "sphinx.ext.mathjax",
    "sphinx.ext.autodoc",
    "sphinx.ext.graphviz",
    "sphinx.ext.extlinks",
    "myst_parser",
    "sphinxext.rediraffe",
]

autodoc_member_order = "bysource"

autodoc_typehints_format = "short"

graphviz_dot_args = [
    "-Gfontname=sans-serif",
    "-Efontname=sans-serif",
    "-Nfontname=sans-serif",
]
graphviz_output_format = "svg"

extlinks = {
    "github_1a5d2c6": (
        "https://github.com/acconeer/acconeer-python-exploration/tree/"
        + "1a5d2c68d1c0b458109818af788ed2b386144644/%s",
        "%s",
    ),
}

rediraffe_redirects = "redirects.txt"
rediraffe_branch = "HEAD~1"
rediraffe_auto_redirect_perc = 95

source_suffix = ".rst"

master_doc = "index"

language = None

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "README.md",
    "how_to_docs.txt",
]

pygments_style = None

html_theme = "sphinx_rtd_theme"

html_css_files = ["css/custom.css"]

html_static_path = ["_static"]

htmlhelp_basename = "acconeer-python-exploration-docs"

latex_elements = {
    "papersize": "a4paper",
    "pointsize": "11pt",
    "fontpkg": r"""
        \usepackage{helvet}
        \renewcommand{\familydefault}{\sfdefault}
    """,
    "tableofcontents": "",
    "sphinxsetup": ",".join(
        [
            r"hmargin={1.2in, 1.2in}",
        ]
    ),
    "preamble": r"""
        \usepackage{titling}
        \usepackage{graphicx}
        \graphicspath{{../../_static/}}
    """,
    "maketitle": r"""
        \vspace*{50mm}
        \begin{center}
            \includegraphics[width=100mm]{logo.pdf}
            \par
            \vspace*{15mm}
            {
                \huge
                \thetitle
            }
        \end{center}
        \newpage
        \tableofcontents
        \newpage
    """,
}

# (startdocname, targetname, title, author, documentclass, toctree_only)
latex_documents = [
    (
        "sensor_introduction",
        "sensor_introduction.tex",
        "Sensor Introduction",
        author,
        "howto",
        False,
    )
]

numfig = True
numfig_format = {"figure": "Figure %s"}
math_eqref_format = "Eq. {number}"
numfig_secnum_depth = 0



import pytest
import requests

from PySide6 import QtCore

from acconeer.exptool.app import GUI
from acconeer.exptool.app.elements.modules import MODULE_INFOS


MOCK_INTERFACE = "Simulated"
LB = QtCore.Qt.LeftButton


def _have_internet_connection():
    try:
        requests.get("https://www.google.com")
        return True
    except requests.exceptions.ConnectionError:
        return False


def _get_gui(qtbot):
    w = GUI(under_test=True)
    qtbot.addWidget(w)
    with qtbot.waitExposed(w):
        w.show()
    return w


@pytest.fixture
def gui(qtbot):
    return _get_gui(qtbot)


@pytest.fixture
def mock_handler_gui(qtbot, mocker):
    """
    Not really interested if the webpage opens in a tab that is the module `webbrowser`'s
    responsibility. We are, however, interested whether the url will resolve (response.ok == True)
    """

    def mock_help_button_handler(self):
        # self == GUI
        url = self.current_module_info.docs_url
        if url:
            response = requests.get(url)
            self.under_test_help_button_response = response
        else:
            self.under_test_help_button_response = None

    mocker.patch("acconeer.exptool.app.GUI.service_help_button_handler", mock_help_button_handler)
    return _get_gui(qtbot)


def test_select_interface(qtbot, gui):
    interfaces = [MOCK_INTERFACE, "SPI", MOCK_INTERFACE]
    for interface in interfaces:
        set_and_check_cb(qtbot, gui.interface_dd, interface)


def test_run_a_session(qtbot, gui):
    assert gui.buttons["connect"].text() == "Connect"
    assert not gui.buttons["start"].isEnabled()
    assert not gui.buttons["stop"].isEnabled()
    assert gui.statusBar().currentMessage() == "Not connected"

    set_and_check_cb(qtbot, gui.module_dd, "Envelope")
    set_and_check_cb(qtbot, gui.interface_dd, MOCK_INTERFACE)
    check_cb(qtbot, gui.module_dd, "Envelope")

    qtbot.mouseClick(gui.buttons["connect"], LB)
    expected_status = "Connected via simulated interface"
    qtbot.waitUntil(lambda: gui.statusBar().currentMessage() == expected_status)
    qtbot.waitUntil(lambda: gui.buttons["connect"].text() == "Disconnect")
    qtbot.waitUntil(lambda: gui.buttons["start"].isEnabled())
    qtbot.waitUntil(lambda: not gui.buttons["stop"].isEnabled())

    qtbot.mouseClick(gui.buttons["start"], LB)
    qtbot.wait(500)
    assert gui.threaded_scan.isRunning()
    assert not gui.buttons["start"].isEnabled()
    assert gui.buttons["stop"].isEnabled()

    with qtbot.waitSignal(gui.sig_scan) as sig:
        qtbot.mouseClick(gui.buttons["stop"], LB)

    assert sig.args[0] == "stop"

    qtbot.waitUntil(lambda: gui.buttons["start"].isEnabled())
    qtbot.waitUntil(lambda: not gui.buttons["stop"].isEnabled())

    qtbot.mouseClick(gui.buttons["connect"], LB)
    qtbot.waitUntil(lambda: gui.buttons["connect"].text() == "Connect")


def test_multi_sensor(qtbot, gui):
    pass  # TODO


def test_start_and_stop_all_modules(qtbot, gui):
    set_and_check_cb(qtbot, gui.interface_dd, MOCK_INTERFACE)
    qtbot.mouseClick(gui.buttons["connect"], LB)

    for module_info in MODULE_INFOS:
        set_and_check_cb(qtbot, gui.module_dd, module_info.label)
        qtbot.wait(200)

        qtbot.mouseClick(gui.buttons["start"], LB)
        qtbot.wait(600)
        qtbot.mouseClick(gui.buttons["stop"], LB)
        qtbot.wait(200)


@pytest.mark.skipif(not _have_internet_connection(), reason="No internet connection")
def test_help_button_with_urls(qtbot, mock_handler_gui):
    for module_info in MODULE_INFOS:
        if not module_info.docs_url:
            continue
        set_and_check_cb(qtbot, mock_handler_gui.interface_dd, MOCK_INTERFACE)
        set_and_check_cb(qtbot, mock_handler_gui.module_dd, module_info.label)
        qtbot.mouseClick(mock_handler_gui.buttons["service_help"], LB)
        resp = mock_handler_gui.under_test_help_button_response
        assert resp and resp.ok
        qtbot.wait(100)


def test_help_button_not_clickable(qtbot, gui):
    for module_info in MODULE_INFOS:
        if module_info.docs_url:
            continue
        set_and_check_cb(qtbot, gui.interface_dd, MOCK_INTERFACE)
        set_and_check_cb(qtbot, gui.module_dd, module_info.label)
        assert not gui.buttons["service_help"].isEnabled()
        qtbot.wait(100)


def connect_and_disconnect(qtbot, gui):
    qtbot.mouseClick(gui.buttons["connect"], LB)
    qtbot.waitUntil(lambda: gui.buttons["connect"].text() == "Disconnect")
    qtbot.mouseClick(gui.buttons["connect"], LB)
    qtbot.waitUntil(lambda: gui.buttons["connect"].text() == "Connect")


def set_cb(cb, text):
    index = cb.findText(text, QtCore.Qt.MatchFixedString)
    assert index >= 0
    cb.setCurrentIndex(index)


def check_cb(qtbot, cb, text, timeout=1000):
    qtbot.waitUntil(lambda: cb.currentText() == text, timeout=timeout)


def set_and_check_cb(qtbot, cb, text, timeout=1000):
    set_cb(cb, text)
    check_cb(qtbot, cb, text, timeout=timeout)



from unittest.mock import Mock

import pytest

from acconeer.exptool.app.elements import helper


@pytest.fixture
def calibration_ui_state() -> helper.CalibrationUiState:
    auto_apply_cb_mock = Mock()
    return helper.CalibrationUiState(auto_apply_cb=auto_apply_cb_mock)


def test_fields_with_newly_instantiated_ui_state(calibration_ui_state):
    assert calibration_ui_state.get_display_text() == ""
    assert not calibration_ui_state.is_display_text_italic
    assert not calibration_ui_state.save_button_enabled
    assert calibration_ui_state.load_button_enabled
    assert not calibration_ui_state.clear_button_enabled
    assert not calibration_ui_state.auto_apply
    assert not calibration_ui_state.apply_button_enabled


@pytest.mark.parametrize(
    "filename,expected",
    [
        ("/home/test/Documents/test.yaml", ".../test.yaml"),
        ("~/test/Documents/test.yaml", ".../test.yaml"),
        ("./test.yaml", "test.yaml"),
    ],
)
def test_display_text_for_file_names(calibration_ui_state, filename, expected):
    calibration_ui_state.source = filename
    assert calibration_ui_state.get_display_text() == expected
    calibration_ui_state.modified = True
    assert calibration_ui_state.get_display_text() == expected + "*"


def test_display_italic(calibration_ui_state):
    assert not calibration_ui_state.is_display_text_italic
    calibration_ui_state.modified = True
    assert calibration_ui_state.is_display_text_italic


def test_display_text_for_session(calibration_ui_state):
    calibration_ui_state.source = "Session"
    assert calibration_ui_state.get_display_text() == "Session"
    calibration_ui_state.modified = True
    assert calibration_ui_state.get_display_text() == "Session*"


def test_save_button_state(calibration_ui_state):
    assert not calibration_ui_state.save_button_enabled
    calibration_ui_state.modified = True
    assert calibration_ui_state.save_button_enabled


def test_load_button_state(calibration_ui_state):
    assert calibration_ui_state.load_button_enabled
    calibration_ui_state.modified = True
    assert not calibration_ui_state.load_button_enabled


def test_apply_button_state(calibration_ui_state):
    assert not calibration_ui_state.apply_button_enabled
    calibration_ui_state.source = "smth"
    assert calibration_ui_state.apply_button_enabled


def test_clear_button_state(calibration_ui_state):
    assert not calibration_ui_state.clear_button_enabled
    calibration_ui_state.source = "Session"
    assert calibration_ui_state.clear_button_enabled


@pytest.mark.parametrize("is_modified", [True, False])
def test_correct_state_after_clear_is_pressed(calibration_ui_state, is_modified):
    calibration_ui_state.source = "a_filename"
    calibration_ui_state.modified = is_modified
    calibration_ui_state.clear()
    assert not calibration_ui_state.save_button_enabled
    assert not calibration_ui_state.clear_button_enabled
    assert calibration_ui_state.load_button_enabled
    assert calibration_ui_state.get_display_text() == ""


def test_save_verb(calibration_ui_state):
    calibration_ui_state.save("test")
    assert calibration_ui_state.source == "test"
    assert not calibration_ui_state.modified


def test_load_verb(calibration_ui_state):
    calibration_ui_state.load("test")
    assert calibration_ui_state.source == "test"
    assert not calibration_ui_state.modified


def test_autoapply_checkbox(calibration_ui_state):
    calibration_ui_state.auto_apply = True
    assert calibration_ui_state.auto_apply
    calibration_ui_state._auto_apply_cb.setChecked.assert_called_with(True)



import time

import numpy as np
import pytest

import acconeer.exptool as et
from acconeer.exptool.a111._clients.json.client import SocketClient
from acconeer.exptool.a111._clients.mock.client import MockClient
from acconeer.exptool.a111._clients.reg.client import SPIClient, UARTClient


@pytest.fixture(scope="module")
def setup(request):
    conn_type, *args = request.param

    if conn_type == "spi":
        client = SPIClient()
        sensor = 1
    elif conn_type == "uart":
        port = args[0] or et.utils.autodetect_serial_port()
        client = UARTClient(port)
        sensor = 1
    elif conn_type == "socket":
        client = SocketClient(args[0])
        sensor = int(args[1])
    elif conn_type == "mock":
        client = MockClient()
        sensor = 1
    else:
        pytest.fail()

    client.connect()
    yield (client, sensor)
    client.disconnect()


def test_run_a_host_driven_session(setup):
    client, sensor = setup

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = sensor
    config.repetition_mode = et.a111.EnvelopeServiceConfig.RepetitionMode.HOST_DRIVEN

    client.start_session(config)
    client.get_next()
    client.stop_session()


def test_run_a_sensor_driven_session(setup):
    client, sensor = setup

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = sensor
    config.repetition_mode = et.a111.EnvelopeServiceConfig.RepetitionMode.SENSOR_DRIVEN
    config.update_rate = 10

    client.start_session(config)
    client.get_next()
    client.stop_session()


def test_run_illegal_config(setup):
    client, sensor = setup

    if isinstance(client, MockClient):
        return

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = sensor
    config.repetition_mode = et.a111.EnvelopeServiceConfig.RepetitionMode.SENSOR_DRIVEN
    config.update_rate = 10
    config.range_interval = [0.2, 2.0]  # too long without stitching

    with pytest.raises(et.a111._clients.base.IllegalConfigError):
        client.setup_session(config)

    with pytest.raises(et.a111._clients.base.SessionSetupError):
        client.setup_session(config, check_config=False)

    config.range_interval = [0.2, 0.4]
    client.start_session(config)
    client.get_next()
    client.stop_session()


@pytest.mark.parametrize("mode", et.a111.Mode)
def test_squeeze(setup, mode):
    client, sensor = setup

    restore_squeeze = client.squeeze

    config = et.a111._configs.MODE_TO_CONFIG_CLASS_MAP[mode]()
    config.sensor = sensor

    client.squeeze = True
    client.start_session(config)
    squeezed_data_info, squeezed_data = client.get_next()
    client.stop_session()

    client.squeeze = False
    client.start_session(config)
    unsqueezed_data_info, unsqueezed_data = client.get_next()
    client.stop_session()

    if mode == et.a111.Mode.SPARSE:
        expected_squeezed_ndim = 2
    else:
        expected_squeezed_ndim = 1

    assert unsqueezed_data.shape == squeezed_data[None, ...].shape
    assert squeezed_data.ndim == expected_squeezed_ndim
    assert unsqueezed_data.ndim == expected_squeezed_ndim + 1
    assert squeezed_data_info.keys() == unsqueezed_data_info[0].keys()

    client.squeeze = restore_squeeze


@pytest.mark.parametrize("mode", et.a111.Mode)
def test_sanity_check_output(setup, mode):
    client, sensor = setup

    config = et.a111._configs.MODE_TO_CONFIG_CLASS_MAP[mode]()
    config.sensor = sensor
    config.range_interval = [0.30, 0.60]  # Important with multiple of 0.06m for sparse

    if mode == et.a111.Mode.SPARSE:
        config.sweeps_per_frame = 10  # Avoid the default value to make sure it bites

    session_info = client.start_session(config)
    data_info, data = client.get_next()
    client.stop_session()

    assert session_info["range_start_m"] == pytest.approx(config.range_start, abs=0.01)
    assert session_info["range_length_m"] == pytest.approx(config.range_length, abs=0.01)
    assert "step_length_m" in session_info

    if mode == et.a111.Mode.POWER_BINS:
        assert "bin_count" in session_info
    else:
        assert "data_length" in session_info

    if mode == et.a111.Mode.SPARSE:
        assert "sweep_rate" in session_info
    else:
        assert "stitch_count" in session_info

    assert type(data_info["data_saturated"]) == bool
    assert type(data_info["missed_data"]) == bool

    if mode != et.a111.Mode.SPARSE:
        assert type(data_info["data_quality_warning"]) == bool

    assert isinstance(data, np.ndarray)

    if mode == et.a111.Mode.POWER_BINS:
        assert data.dtype == float
        size = session_info["bin_count"]
        assert data.shape == (size,)
        assert 1 < size < 10
    elif mode == et.a111.Mode.ENVELOPE:
        assert data.dtype == float
        size = session_info["data_length"]
        assert data.shape == (size,)
        assert size == pytest.approx(config.range_length / 0.06 * 124, abs=10)
    elif mode == et.a111.Mode.IQ:
        assert data.dtype == complex
        size = session_info["data_length"]
        assert data.shape == (size,)
        assert size == pytest.approx(config.range_length / 0.06 * 124, abs=10)
    elif mode == et.a111.Mode.SPARSE:
        assert data.dtype == float
        data_length = session_info["data_length"]
        num_depths = data_length // config.sweeps_per_frame
        assert num_depths * config.sweeps_per_frame == data_length
        assert data.shape == (config.sweeps_per_frame, num_depths)
        assert 1 < config.sweeps_per_frame <= 64
        assert abs(num_depths - int(float(config.range_length / 0.06))) <= 2
    else:
        pytest.fail("test does not cover all modes")


@pytest.mark.parametrize("mode", et.a111.Mode)
def test_downsampling_factor(setup, mode):
    client, sensor = setup

    if mode == et.a111.Mode.POWER_BINS:
        return

    config = et.a111._configs.MODE_TO_CONFIG_CLASS_MAP[mode]()
    config.sensor = sensor
    config.range_interval = [0.30, 0.60]

    size = None

    for df in (1, 2):
        config.downsampling_factor = df

        session_info = client.start_session(config)
        data_info, data = client.get_next()
        client.stop_session()

        step_length = session_info["step_length_m"]

        if mode == et.a111.Mode.SPARSE:
            expected = df * 0.06
        else:  # Envelope, IQ
            expected = df * 0.48e-3

        assert step_length == pytest.approx(expected, rel=0.1)

        if size is None:  # df == 1
            size = session_info["data_length"]
        else:  # df == 2, size is assuming df == 1
            assert abs(size - 2 * session_info["data_length"]) <= 2


def test_repetition_mode(setup):
    client, sensor = setup

    # TODO Test not stable for exploration server
    if isinstance(client, SocketClient):
        pytest.skip("Skip socket client")

    def measure(config):
        client.start_session(config, check_config=False)
        client.get_next()
        t0 = time.time()

        missed = False
        n = 5

        for _ in range(n):
            info, data = client.get_next()
            if info["missed_data"]:
                missed = True

        t1 = time.time()
        client.stop_session()
        dt = (t1 - t0) / n
        return (dt, missed)

    config = et.a111.SparseServiceConfig()
    config.sensor = sensor
    config.range_interval = [0.3, 0.36]
    config.sweeps_per_frame = 50
    config.sweep_rate = 1e3

    nominal_f = config.sweep_rate / config.sweeps_per_frame
    nominal_dt = 1.0 / nominal_f

    # on demand / host driven
    config.repetition_mode = et.a111.SparseServiceConfig.RepetitionMode.HOST_DRIVEN

    # no rate limit
    config.update_rate = None
    dt, missed = measure(config)
    assert not missed
    assert dt == pytest.approx(nominal_dt, rel=0.15)

    # ok rate limit
    config.update_rate = 0.5 / nominal_dt
    dt, missed = measure(config)
    assert not missed
    assert dt == pytest.approx(nominal_dt * 2.0, rel=0.15)

    # too high rate limit
    config.update_rate = 2.0 / nominal_dt
    dt, missed = measure(config)

    if isinstance(client, SocketClient):  # TODO
        assert missed

    assert dt == pytest.approx(nominal_dt, rel=0.15)

    # streaming / sensor driven
    config.repetition_mode = et.a111.SparseServiceConfig.RepetitionMode.SENSOR_DRIVEN

    # ok rate
    config.update_rate = 0.5 / nominal_dt
    dt, missed = measure(config)
    assert not missed
    assert dt == pytest.approx(nominal_dt * 2.0, rel=0.15)

    # too high rate
    config.update_rate = 2.0 / nominal_dt
    dt, missed = measure(config)
    assert missed



def pytest_addoption(parser):
    parser.addoption(
        "--uart",
        dest="uart",
        metavar="port",
        nargs="?",
        const="",
    )

    parser.addoption(
        "--spi",
        dest="spi",
        action="store_true",
    )

    parser.addoption(
        "--socket",
        dest="socket",
        metavar="socket_args",
        nargs=2,
    )

    parser.addoption(
        "--mock",
        dest="mock",
        action="store_true",
    )


def ids_fun(setup):
    try:
        return setup[0]
    except Exception:
        return ""


def pytest_generate_tests(metafunc):
    FIXTURE_NAME = "setup"

    if FIXTURE_NAME in metafunc.fixturenames:
        params = []

        uart_port = metafunc.config.getoption("uart")
        if uart_port is not None:
            params.append(("uart", uart_port))

        spi = metafunc.config.getoption("spi")
        if spi:
            params.append(("spi",))

        socket = metafunc.config.getoption("socket")
        if socket is not None:
            params.append(("socket", *socket))

        mock = metafunc.config.getoption("mock")
        if mock:
            params.append(("mock",))

        metafunc.parametrize(FIXTURE_NAME, params, indirect=True, ids=ids_fun)



import tempfile
from pathlib import Path

import attrs
import h5py
import numpy as np
import pytest
import yaml

import acconeer.exptool as et
from acconeer.exptool.a111.algo.obstacle_detection import ProcessingConfiguration, Processor
from acconeer.exptool.a111.algo.obstacle_detection.calibration import ObstacleDetectionCalibration


HERE = Path(__file__).parent
TEST_KEYS = [
    "env_ampl",
    "peak_idx",
    "angle",
    "velocity",
    "peaks_found",
]
PARAMETER_SETS = [
    {},
]


def get_bg_parameters(parameter_set=None):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()

    if parameter_set is not None:
        for k, v in parameter_set.items():
            processing_config[k]["value"] = v

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
    )

    calibrations = []

    for data_info, data in input_record:
        if data.shape[0] == 1:
            result = processor.process(data.squeeze(0), data_info[0])
        else:
            result = processor.process(data, data_info[0])
        if "new_calibration" in result:
            calibrations.append(result["new_calibration"])

    assert len(calibrations) == 1

    return attrs.asdict(calibrations[0])


def get_output_for_calib(parameter_set=None):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()
    calibration = ObstacleDetectionCalibration.load(HERE / "obstacle_bg_params_dump.yaml")

    if parameter_set is not None:
        for k, v in parameter_set.items():
            processing_config[k]["value"] = v

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
        calibration,
    )

    output = {k: [] for k in TEST_KEYS}

    for data_info, data in input_record:

        if data.shape[0] == 1:
            result = processor.process(data.squeeze(0), data_info[0])
        else:
            result = processor.process(data, data_info[0])

        assert "new_calibration" not in result

        for k in TEST_KEYS:
            output[k].append(result[k])

    # Explicit `dtype=float` makes the conversion `None` -> `np.nan`.
    return {k: np.array(v, dtype=float) for k, v in output.items()}


def get_output(parameter_set=None):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()

    if parameter_set is not None:
        for k, v in parameter_set.items():
            processing_config[k]["value"] = v

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
    )

    output = {k: [] for k in TEST_KEYS}

    for data_info, data in input_record:
        if data.shape[0] == 1:
            result = processor.process(data.squeeze(0), data_info[0])
        else:
            result = processor.process(data, data_info[0])

        for k in TEST_KEYS:
            output[k].append(result[k])

    # Explicit `dtype=float` makes the conversion `None` -> `np.nan`.
    return {k: np.array(v, dtype=float) for k, v in output.items()}


def save_output(file, output):
    with h5py.File(file, "w") as f:
        for k in TEST_KEYS:
            try:
                f.create_dataset(name=k, data=output[k], track_times=False, compression="gzip")
            except TypeError as te:
                raise TypeError(
                    f"Could not create dataset with name: {k}, data={output[k]}"
                ) from te


def load_output(file):
    output = {}

    with h5py.File(file, "r") as f:
        for k in TEST_KEYS:
            output[k] = f[k][()]

    return output


def compare_dicts(expected, actual, keys={}, exact=False) -> bool:
    if not keys:
        keys = expected.keys()
        assert expected.keys() == actual.keys()

    result = True
    for key in keys:
        expected_val = expected[key]
        actual_val = actual[key]

        if exact:
            result = result and np.all(expected_val == actual_val)
        else:
            result = result and np.allclose(expected_val, actual_val, equal_nan=True)

    return result


def compare_output(expected, actual, exact=False):
    assert compare_dicts(expected, actual, TEST_KEYS, exact)


def path_for_parameter_set(parameter_set):
    if parameter_set:
        l = sorted(parameter_set.items())
        suffix = "_".join(f"{k}-{v}" for k, v in l)
    else:
        suffix = "default"

    return HERE / f"output_{suffix}.h5"


def test_load_save_compare():
    temp_file = tempfile.TemporaryFile()

    saved_output = get_output()

    save_output(temp_file, saved_output)
    loaded_output = load_output(temp_file)

    compare_output(saved_output, loaded_output)


def test_path_for_parameter_set():
    assert path_for_parameter_set({"foo": "bar"}) == (HERE / "output_foo-bar.h5")


@pytest.mark.parametrize("parameter_set", PARAMETER_SETS)
def test_processor_against_reference(parameter_set):
    with open(path_for_parameter_set(parameter_set), "rb") as f:
        expected = load_output(f)

    actual = get_output(parameter_set)
    compare_output(expected, actual)


@pytest.mark.parametrize("parameter_set", PARAMETER_SETS)
def test_dumped_parameters(parameter_set):
    actual_dict = get_bg_parameters(parameter_set)

    with open(HERE / "obstacle_bg_params_dump.yaml") as stream:
        expected_dict = yaml.safe_load(stream)

    assert compare_dicts(expected_dict, actual_dict)


@pytest.mark.parametrize("parameter_set", PARAMETER_SETS)
def test_wrong_dumped_parameters(parameter_set):
    actual_dict = get_bg_parameters(parameter_set)

    with open(HERE / "obstacle_bg_params_dump_wrong.yaml") as stream:
        wrong_dict = yaml.safe_load(stream)

    assert actual_dict.keys() == wrong_dict.keys()

    assert not compare_dicts(actual_dict, wrong_dict)


@pytest.mark.parametrize("parameter_set", PARAMETER_SETS)
def test_setting_calibration_yeilds_expected_results(parameter_set):
    path = path_for_parameter_set(parameter_set).as_posix().replace(".h5", "_calib.h5")
    with open(path, "rb") as f:
        expected = load_output(f)

    actual = get_output_for_calib(parameter_set)
    assert compare_dicts(actual, expected)


@pytest.mark.parametrize("parameter_set", PARAMETER_SETS)
def test_calibration_instantiated_processor_does_not_return_new_calibration(parameter_set):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()
    calibration = ObstacleDetectionCalibration.load(HERE / "obstacle_bg_params_dump.yaml")

    for k, v in parameter_set.items():
        processing_config[k]["value"] = v

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
        calibration,
    )

    for data_info, data in input_record:
        result = processor.process(data, data_info[0])
        assert "new_calibration" not in result


@pytest.mark.parametrize("parameter_set", PARAMETER_SETS)
def test_update_calibration_actually_updates(parameter_set):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()
    calibration = ObstacleDetectionCalibration.load(HERE / "obstacle_bg_params_dump.yaml")
    wrong_calibration = ObstacleDetectionCalibration.load(
        HERE / "obstacle_bg_params_dump_wrong.yaml"
    )

    if parameter_set is not None:
        for k, v in parameter_set.items():
            processing_config[k]["value"] = v

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
    )

    for _ in range(10):
        processor.update_calibration(wrong_calibration)
    processor.update_calibration(calibration)

    output = {k: [] for k in TEST_KEYS}

    for data_info, data in input_record:

        if data.shape[0] == 1:
            result = processor.process(data.squeeze(0), data_info[0])
        else:
            result = processor.process(data, data_info[0])

        assert "new_calibration" not in result

        for k in TEST_KEYS:
            output[k].append(result[k])

    # Explicit `dtype=float` makes the conversion `None` -> `np.nan`.
    actual = {k: np.array(v, dtype=float) for k, v in output.items()}

    path = path_for_parameter_set(parameter_set).as_posix().replace(".h5", "_calib.h5")
    with open(path, "rb") as f:
        expected = load_output(f)

    assert compare_dicts(actual, expected)


@pytest.mark.parametrize("parameter_set", PARAMETER_SETS)
def test_preloaded_bg_params(parameter_set):
    actual = get_output_for_calib(parameter_set)

    path = path_for_parameter_set(parameter_set).as_posix().replace(".h5", "_calib.h5")
    with open(path, "rb") as f:
        expected = load_output(f)

    assert compare_dicts(actual, expected)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    sp = subparsers.add_parser("save")

    args = parser.parse_args()

    if args.command == "save":
        for parameter_set in PARAMETER_SETS:
            output = get_output(parameter_set)
            save_output(path_for_parameter_set(parameter_set), output)

        for parameter_set in PARAMETER_SETS:
            output = get_output_for_calib(parameter_set)
            save_output(
                path_for_parameter_set(parameter_set).as_posix().replace(".h5", "_calib.h5"),
                output,
            )
    else:
        raise RuntimeError



import functools

import pytest

import acconeer.exptool as et
from acconeer.exptool import structs
from acconeer.exptool.a111 import _configs, _modes


def is_test(member_name, modules):
    """the `is` operator is reference equality, while `==` is for "member equality"
    i.e. 2 different objects have the same contents.

    We want to verify that one member `is` the same across namespaces.
    """
    # Gets the same member from different modules
    members = [getattr(module, member_name) for module in modules]

    # if `a is b`, `a` is safe as a reduction.
    # if `a is not b` then None will be the reduction
    # (and also the final reduction)
    def _is(a, b):
        return a if (a is b) else None

    return functools.reduce(_is, members) is not None


@pytest.mark.parametrize(
    "config_type",
    [
        "EnvelopeServiceConfig",
        "IQServiceConfig",
        "PowerBinServiceConfig",
        "SparseServiceConfig",
    ],
)
def test_top_module_configs(config_type):
    assert et.a111._configs == _configs
    assert is_test(config_type, [et.a111, _configs, et.a111._configs])


@pytest.mark.parametrize("structs_member", ["configbase"])
def test_top_module_structs(structs_member):
    assert is_test(structs_member, [et, structs, et.structs])


def test_top_module_mode():
    assert is_test("Mode", [et.a111, _modes])



import tempfile
from pathlib import Path

import h5py
import numpy as np

import acconeer.exptool as et
from acconeer.exptool.a111.algo.button_press_sparse._processor import (
    ProcessingConfiguration,
    Processor,
)


HERE = Path(__file__).parent

TEST_KEYS = ["signal", "detection"]
PARAMETER_SETS = [
    {},
]


def get_output(parameter_set=None):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()

    if parameter_set is not None:
        for k, v in parameter_set.items():
            setattr(processing_config, k, v)

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
    )

    output = {k: [] for k in TEST_KEYS}

    for data_info, data in input_record:
        result = processor.process(data.squeeze(0), data_info[0])

        for k in TEST_KEYS:
            output[k].append(result[k])

    return {k: np.array(v) for k, v in output.items()}


def save_output(file, output):
    with h5py.File(file, "w") as f:
        for k in TEST_KEYS:
            f.create_dataset(name=k, data=output[k], track_times=False)


def load_output(file):
    output = {}

    with h5py.File(file, "r") as f:
        for k in TEST_KEYS:
            output[k] = f[k][()]

    return output


def compare_output(expected, actual, exact=False):
    for k in TEST_KEYS:
        expected_arr = expected[k]
        actual_arr = actual[k]

        if exact:
            assert np.all(expected_arr == actual_arr)
        else:
            assert np.all(np.isclose(expected_arr, actual_arr))


def path_for_parameter_set(parameter_set):
    if parameter_set:
        l = sorted(parameter_set.items())
        suffix = "_".join(f"{k}-{v}" for k, v in l)
    else:
        suffix = "default"

    return HERE / f"output_{suffix}.h5"


def test_load_save_compare():
    temp_file = tempfile.TemporaryFile()

    saved_output = get_output()

    save_output(temp_file, saved_output)
    loaded_output = load_output(temp_file)

    compare_output(saved_output, loaded_output, exact=True)


def test_path_for_parameter_set():
    assert path_for_parameter_set({"foo": "bar"}) == (HERE / "output_foo-bar.h5")


def test_processor_against_reference():
    for parameter_set in PARAMETER_SETS:
        with open(path_for_parameter_set(parameter_set), "rb") as f:
            expected = load_output(f)

        actual = get_output(parameter_set)
        compare_output(expected, actual)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    sp = subparsers.add_parser("save")

    args = parser.parse_args()

    if args.command == "save":
        for parameter_set in PARAMETER_SETS:
            output = get_output(parameter_set)
            save_output(path_for_parameter_set(parameter_set), output)
    else:
        raise RuntimeError



import sys

import pytest
import serial
from serial.tools.list_ports_common import ListPortInfo

from acconeer.exptool import utils


pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Windows unsupported")


@pytest.fixture
def mock_comports():
    mock_port_attrs = [
        {
            "device": "/dev/ttyS4",
        },
        {  # XB112 ~ as given from serial.tools.list_ports.comports
            "device": "/dev/ttyUSB0",
            "description": "XB112",
            "manufacturer": "Acconeer AB",
            "product": "XB112",
            "interface": None,
        },
        {  # XE132 ~ as given ...
            "device": "/dev/ttyUSB1",
            "description": "Acconeer XE132 - Enhanced Com Port",
            "manufacturer": "Silicon Labs",
            "product": "Acconeer XE132",
            "interface": "Enhanced Com Port",
        },
        {  # XE132 ~ as given ...
            "device": "/dev/ttyUSB2",
            "description": "Acconeer XE132 - Standard Com Port",
            "manufacturer": "Silicon Labs",
            "product": "Acconeer XE132",
            "interface": "Standard Com Port",
        },
        {
            "device": "/dev/ttyS5",
        },
        {
            "device": "/dev/ttyS6",
        },
        {
            "device": "/dev/ttyUSB3",
            "description": "Linux Foundation 1.1 root hub ",
        },
    ]
    mock_ports = []
    for mpp in mock_port_attrs:
        lpi = ListPortInfo("")
        for k, v in mpp.items():
            setattr(lpi, k, v)
        mock_ports.append(lpi)
    return mock_ports


def _only_acconeer_devices(tagged_port_infos):
    return [tagged for tagged in tagged_port_infos if tagged[1]]


def test_port_tagging_keeps_all_ports(mock_comports):
    tagged_ports = utils.tag_serial_ports(mock_comports)
    assert len(mock_comports) == len(tagged_ports)


@pytest.mark.parametrize("pyserial_version", ["3.4", "3.5"])
def test_port_tagging_finds_all_acconeer_devices(mock_comports, pyserial_version, mocker):
    expected = 2
    mocker.patch.object(serial, "__version__", pyserial_version)
    if pyserial_version == "3.4":
        # Emulate behaviour of pyserial==3.4 regardless of actual version
        mock_comports[2].interface = None
        mock_comports[3].interface = None
        mock_comports[2].apply_usb_info()
        mock_comports[3].apply_usb_info()
        expected = 3
    tagged_ports = utils.tag_serial_ports(mock_comports)
    only_acconeer = _only_acconeer_devices(tagged_ports)
    assert len(only_acconeer) == expected



from typing import List, Type, Union

import pytest

from acconeer.exptool.a111._clients import links
from acconeer.exptool.a111._clients.client_factory import ClientFactory
from acconeer.exptool.a111._clients.json.client import SocketClient
from acconeer.exptool.a111._clients.reg.client import SPIClient, UARTClient


def resolve_attribute_path(obj, attribute_path):
    """
    `attribute_path` should be a path that can be appended to `obj`

    E.g.:

    obj.link.protocol.foo =>
        (obj, attribute_path) = (obj, "link.protocol.foo")
    """
    for attribute in attribute_path.split("."):
        if not attribute:
            continue
        obj = getattr(obj, attribute)
    return obj


def assert_client_types(obj, attribute_path, types: Union[List[Type], Type]):
    """
    `types` is a single type or a list of types, where the attribute
    at `attribute_path` needs to be atleast one of the types.
    """
    obj = resolve_attribute_path(obj, attribute_path)

    if isinstance(types, list):
        assert any([type(obj) == t for t in list(types)])
    else:
        type_ = types
        assert type(obj) == type_


def assert_client_values(obj, attribute_path, value):
    """Analogue to `assert_client_types` but value has no options, hence no list."""
    obj = resolve_attribute_path(obj, attribute_path)
    assert obj == value


def test_conflicting_keyword_arguments_raises_value_error():
    with pytest.raises(ValueError):
        ClientFactory.from_kwargs(**dict(serial_port="port", host="host"))


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(link="socket"),  # only socket no host
        dict(protocol="b", link="socket", host="host"),  # socket, bad protocol
        dict(protocol="b", link="uart", serial_port="port"),  # uart, bad protocol
        dict(link="blablabla"),  # bad link
    ],
)
def test_supplying_weird_arguments_raises_value_error(kwargs):
    with pytest.raises(ValueError):
        ClientFactory.from_kwargs(**kwargs)


@pytest.mark.parametrize(
    "mock_autodetect",
    [[], [("port_a", "tag_a"), ("port_b", "tag_b")]],  # no ports detected  # multiple detected
)
def test_single_acconeer_port_fails(mock_autodetect):
    with pytest.raises(ValueError):
        ClientFactory._get_single_acconeer_module_port(mock_autodetect)


def test_single_acconeer_port_success():
    assert ClientFactory._get_single_acconeer_module_port([("port", "tag")]) == "port"


@pytest.mark.parametrize(
    "kwargs,expected_links_enum",
    [
        (dict(protocol="module"), "uart"),  # Only protocol => infer uart
        (dict(host="mock-host"), "socket"),  # Only host => infer socket
        ({}, "uart"),  # No args -> Infer uart
    ],
)
def test_try_infer_link_from_kwargs(kwargs, expected_links_enum):
    assert ClientFactory._try_infer_link_from_kwargs(**kwargs) == expected_links_enum


@pytest.mark.parametrize("kwargs,expected_protocols_enum", [])  # Nothing here yet
def test_try_infer_protocol_from_kwargs(kwargs, expected_protocols_enum):
    assert ClientFactory._try_infer_protocol_from_kwargs(**kwargs) == expected_protocols_enum


@pytest.mark.parametrize(
    "kwargs,expected_types,expected_values",
    [
        # Only host => infer SocketClient
        (
            dict(host="host"),
            [("", SocketClient), ("_link", links.SocketLink)],
            [("_link._host", "host")],
        ),
        # Exploration over socket
        (
            dict(protocol="exploration", link="socket", host="host"),
            [("", SocketClient), ("_link", links.SocketLink)],
            [("_link._host", "host")],
        ),
        # Exploration over UART
        (
            dict(protocol="exploration", link="uart", serial_port="port"),
            [("", SocketClient), ("_link", links.ExploreSerialLink)],
            [("_link._port", "port")],
        ),
        # Streaming over socket
        (
            dict(protocol="streaming", link="socket", host="host"),
            [("", SocketClient), ("_link", links.SocketLink)],
            [("_link._host", "host")],
        ),
        # Module over UART
        (
            dict(protocol="module", link="uart", serial_port="port"),
            [
                ("", UARTClient),
                ("_link", [links.SerialLink, links.SerialProcessLink]),
            ],
            [("_link._port", "port")],
        ),
        # Module over SPI
        (
            dict(protocol="module", link="spi"),
            [("", SPIClient)],
            [],
        ),
    ],
)
def test_factory(kwargs, expected_types, expected_values):
    client = ClientFactory.from_kwargs(**kwargs)
    for expected_value in expected_values:
        assert_client_values(client, *expected_value)
    for expected_type in expected_types:
        assert_client_types(client, *expected_type)



import sys
from itertools import chain
from pathlib import Path

import pytest

import acconeer.exptool as et


HERE = Path(__file__).parent
path = (HERE / ".." / ".." / "utils").resolve()
sys.path.append(path.as_posix())

from convert_to_csv import record_to_csv  # noqa: E402


@pytest.mark.parametrize("test_file", chain(HERE.glob("**/*.h5"), HERE.glob("**/*.npz")))
def test_csv_conversion_is_exact(test_file):
    # The idea is to test the csv conversion corresponds exactly to the data file.
    # Aimed to catch rounding errors and flipped cols/rows.
    record = et.recording.load(test_file)
    if record.mode == et.Mode.SPARSE:
        pytest.skip("CSV-ifying of sparse data is not supported at this moment.")

    data = record.data.squeeze()
    assert data.ndim == 2

    csv_table = record_to_csv(record)
    csv_table_sac = record_to_csv(record, sweep_as_column=True)

    assert data.shape == csv_table.shape
    assert data.T.shape == csv_table_sac.shape

    for row in range(data.shape[0]):
        for col in range(data.shape[1]):
            assert data[row, col] == complex(csv_table[row, col])
            assert data[row, col] == complex(csv_table_sac[col, row])



# my_algorithm_module.py

from enum import Enum

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool.a111.algo import ModuleFamily, ModuleInfo


class MySensorConfig(et.a111.EnvelopeServiceConfig):
    """Defines default sensor config and service to use"""

    def __init__(self):
        super().__init__()
        self.profile = et.a111.EnvelopeServiceConfig.Profile.PROFILE_1
        self.range_interval = [0.1, 0.5]  # in meters
        self.running_average_factor = 0.01
        self.maximize_signal_attenuation = True
        self.update_rate = 60
        self.gain = 0.2
        self.repetition_mode = et.a111.EnvelopeServiceConfig.RepetitionMode.SENSOR_DRIVEN


class MyProcessingConfiguration(et.configbase.ProcessingConfig):
    """
    Define configuration options for plotting and processing.
    The GUI will populate buttons and sliders for
    all parameters defined here. Check the other detectors for examples!
    """

    VERSION = 1

    class PlotColors(Enum):
        ACCONEER_BLUE = "#38bff0"
        WHITE = "#ffffff"
        BLACK = "#000000"
        PINK = "#f280a1"

    plot_color = et.configbase.EnumParameter(
        label="Plot color",
        enum=PlotColors,
        default_value=PlotColors.ACCONEER_BLUE,
        updateable=True,
        order=10,
        help="What color the plot graph should be",
    )

    scale = et.configbase.FloatParameter(
        label="Scale",
        default_value=1.0,
        decimals=3,
        limits=[0.001, 1.0],
        updateable=True,
        order=20,
        help="Allows you to scale the incoming envelope by a factor",
    )


class MyNewProcessor:
    """Processor class, which should do all the processing in the example."""

    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.scale = processing_config.scale

    def update_processing_config(self, processing_config):
        """This function is called when you change sliders or values in the GUI"""

        self.scale = processing_config.scale

    def process(self, data, data_info):
        """
        This function is called every frame and should return the `dict` out_data.
        """

        scaled_data = self.scale * data
        out_data = {"scaled_data": scaled_data}
        return out_data


class MyPGUpdater:
    """This class does all the plotting."""

    def __init__(self, sensor_config, processing_config, session_info):
        self.plot_color = processing_config.plot_color.value
        self.data_length = session_info["data_length"]

    def setup(self, win):
        """
        This function sets up all graphs and plots. Check the other
        detectors and examples to see how to initialize different
        types of graphs and plots!
        """
        win.setWindowTitle("My new example")

        self.my_plot = win.addPlot(title="My Plot")
        self.my_plot.setMenuEnabled(False)
        self.my_plot.setMouseEnabled(x=False, y=False)
        self.my_plot.hideButtons()
        self.my_plot.addLegend()
        self.my_plot.showGrid(x=True, y=True)
        self.my_plot.setXRange(0, self.data_length)
        self.my_plot.setYRange(0, 100)

        self.my_plot_curve = self.my_plot.plot(
            pen=pg.mkPen(self.plot_color, width=2),
            name="Envelope signal",
        )

    def update_processing_config(self, processing_config=None):
        """This function is called when you change sliders or values in the GUI"""

        self.plot_color = processing_config.plot_color.value
        self.my_plot_curve.setPen(self.plot_color, width=2)

    def update(self, out_data):
        """
        This function is called each frame and receives the dict `out_data` from `MyProcessor`.
        """
        data_from_my_processor = out_data["scaled_data"]
        self.my_plot_curve.setData(data_from_my_processor)


my_module_info = ModuleInfo(
    key="my_algorithm_module",
    label="My Algorithm Module",
    module_family=ModuleFamily.EXAMPLE,
    multi_sensor=False,
    docs_url=None,
    pg_updater=MyPGUpdater,
    processing_config_class=MyProcessingConfiguration,
    sensor_config_class=MySensorConfig,
    processor=MyNewProcessor,
)



import pytest

from acconeer.exptool.a111 import Mode, get_mode


def test_special_cases():
    with pytest.raises(ValueError):
        get_mode("does-not-exist")

    with pytest.raises(ValueError):
        get_mode(object())

    assert get_mode(None) is None


@pytest.mark.parametrize("mode", Mode)
def test_positive(mode):
    assert get_mode(mode) == mode

    assert get_mode(mode.name.lower()) == mode
    assert get_mode(mode.name.upper()) == mode
    assert get_mode(mode.name.title()) == mode

    assert get_mode(mode.value.lower()) == mode
    assert get_mode(mode.value.upper()) == mode
    assert get_mode(mode.value.title()) == mode



import acconeer.exptool.a111._clients.reg.protocol as ptcl
from acconeer.exptool.a111 import Mode
from acconeer.exptool.a111._clients.reg import regmap


unp_reg_val = ptcl.RegVal(2, b"\x03\x00\x00\x00")
unp_reg_read_res = ptcl.RegReadResponse(unp_reg_val)
pkd_reg_read_res_segment = b"\x02\x03\x00\x00\x00"
pkd_reg_read_res_packet = bytearray([ptcl.REG_READ_RESPONSE]) + pkd_reg_read_res_segment
pkd_reg_read_res_frame = (
    bytearray([ptcl.START_MARKER])
    + b"\x05\x00"
    + pkd_reg_read_res_packet
    + bytearray([ptcl.END_MARKER])
)

unp_reg_write_req = ptcl.RegWriteRequest(unp_reg_val)
pkd_reg_write_req_packet = bytearray()
pkd_reg_write_req_packet.append(ptcl.REG_WRITE_REQUEST)
pkd_reg_write_req_packet.append(unp_reg_write_req.reg_val.addr)
pkd_reg_write_req_packet.extend(unp_reg_write_req.reg_val.val)
pkd_reg_write_req_frame = bytearray()
pkd_reg_write_req_frame.append(ptcl.START_MARKER)
pkd_reg_write_req_frame.extend(b"\x05\x00")  # len
pkd_reg_write_req_frame.extend(pkd_reg_write_req_packet)
pkd_reg_write_req_frame.append(ptcl.END_MARKER)


def test_unpack_packet():
    unpacked = ptcl.unpack_packet(pkd_reg_read_res_packet)
    assert unpacked == unp_reg_read_res


def test_unpack_reg_read_res_segment():
    unpacked = ptcl.unpack_reg_read_res_segment(pkd_reg_read_res_segment)
    assert unpacked == unp_reg_read_res


def test_unpack_stream_data_segment():
    reg = regmap.get_reg("run_factor", Mode.ENVELOPE)
    rv_addr = reg.addr
    rv_enc_val = reg.encode(123)
    rvs = [ptcl.RegVal(rv_addr, rv_enc_val)]
    buffer = bytearray(b"\x12\x34\x56")
    unp_stream_data = ptcl.StreamData(rvs, buffer)

    pkd_stream_data_segment = bytearray()
    pkd_stream_data_segment.append(ptcl.STREAM_BUFFER)
    pkd_stream_data_segment.extend(b"\x03\x00")
    pkd_stream_data_segment.extend(buffer)
    pkd_stream_data_segment.append(ptcl.STREAM_RESULT_INFO)
    pkd_stream_data_segment.extend(b"\x05\x00")
    pkd_stream_data_segment.append(rv_addr)
    pkd_stream_data_segment.extend(rv_enc_val)

    unpacked = ptcl.unpack_stream_data_segment(pkd_stream_data_segment)
    assert unpacked == unp_stream_data


def test_pack_packet():
    packed = ptcl.pack_packet(unp_reg_write_req)
    assert packed == pkd_reg_write_req_packet


def test_extract_packet_from_frame():
    packet = ptcl.extract_packet_from_frame(pkd_reg_read_res_frame)
    assert packet == pkd_reg_read_res_packet


def test_insert_packet_into_frame():
    frame = ptcl.insert_packet_into_frame(unp_reg_write_req)
    assert frame == pkd_reg_write_req_frame



import tempfile
from contextlib import nullcontext as does_not_raise
from enum import Enum
from pathlib import Path

import h5py
import numpy as np
import pytest

import acconeer.exptool as et
from acconeer.exptool.a111.algo.distance_detector._processor import (
    ProcessingConfiguration,
    Processor,
)
from acconeer.exptool.a111.algo.distance_detector.calibration import DistanceDetectorCalibration


HERE = Path(__file__).parent

TEST_KEYS = [
    "threshold",
    "above_thres_hist_dist",
    "found_peaks",
]
PARAMETER_SETS = [
    {
        "threshold_type": ProcessingConfiguration.ThresholdType.CFAR,
        "cfar_sensitivity": 0.7,  # input produces no detections for default (=0.5)
    },
]
PARAMETER_SETS_CALIBRATION = [
    {"threshold_type": ProcessingConfiguration.ThresholdType.RECORDED},
]


def list_of_lists_to_matrix(list_of_lists):
    """
    Some of the parameters returns list of lists (with dynamic length). In
    order for these to be saved to a h5 file, they are put in a matrix with
    "nan-terminated" rows.
    """
    rows = len(list_of_lists)
    columns = max(len(list) for list in list_of_lists)
    matrix = np.full((rows, columns), fill_value=np.nan)

    for i, list in enumerate(list_of_lists):
        matrix[i, 0 : len(list)] = list

    return matrix


def get_output(parameter_set=None, processor_modifier=None):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()

    if parameter_set is not None:
        for k, v in parameter_set.items():
            setattr(processing_config, k, v)

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
    )

    if processor_modifier:
        processor = processor_modifier(processor)

    output = {k: [] for k in TEST_KEYS}

    for data_info, data in input_record:
        result = processor.process(data.squeeze(0), data_info[0])

        for k in TEST_KEYS:
            if result[k] is None:
                output[k].append([])
            else:
                output[k].append(result[k])

    return {key: list_of_lists_to_matrix(data) for key, data in output.items()}


def get_output_precalibrated(parameter_set=None):
    if parameter_set.get("threshold_type") != ProcessingConfiguration.ThresholdType.RECORDED:
        pytest.skip(f'Test is N/A with threshold_type={parameter_set.get("threshold_type")}')

    def calibration_updater(processor):
        calibration = DistanceDetectorCalibration.load(HERE / "calibration.npy")
        processor.update_calibration(calibration)
        return processor

    return get_output(parameter_set=parameter_set, processor_modifier=calibration_updater)


def get_calibration(parameter_set=None):
    input_record = et.a111.recording.load(HERE / "input.h5")

    processing_config = ProcessingConfiguration()

    if parameter_set is not None:
        for k, v in parameter_set.items():
            setattr(processing_config, k, v)

    processor = Processor(
        input_record.sensor_config,
        processing_config,
        input_record.session_info,
    )

    for data_info, data in input_record:
        output = processor.process(data.squeeze(0), data_info[0])
        if "new_calibration" in output:
            return output["new_calibration"]
    assert False


def save_output(file, output):
    with h5py.File(file, "w") as f:
        for k in TEST_KEYS:
            try:
                f.create_dataset(name=k, data=output[k], track_times=False)
            except TypeError as te:
                raise TypeError(f'Key "{k}" is probably of wrong type ({type(output[k])})') from te


def load_output(file):
    output = {}

    with h5py.File(file, "r") as f:
        for k in TEST_KEYS:
            output[k] = f[k][()]

    return output


def compare_output(expected, actual, exact=False):
    for k in TEST_KEYS:
        expected_arr = expected[k]
        actual_arr = actual[k]

        if exact:
            np.testing.assert_array_equal(expected_arr, actual_arr)
        else:
            np.testing.assert_allclose(expected_arr, actual_arr)


def path_for_parameter_set(parameter_set, calibration=False):
    if parameter_set:
        l = sorted(parameter_set.items())

        for i, (key, maybe_enum) in enumerate(l):
            if isinstance(maybe_enum, Enum):
                enum_member = maybe_enum
                l[i] = (key, enum_member.name.lower())

        suffix = "_".join(f"{k}-{v}" for k, v in l)
    else:
        suffix = "default"

    if calibration:
        suffix += "_calibration"

    return HERE / f"output_{suffix}.h5"


def test_load_save_compare():
    temp_file = tempfile.TemporaryFile()

    saved_output = get_output()

    save_output(temp_file, saved_output)
    loaded_output = load_output(temp_file)

    compare_output(saved_output, loaded_output, exact=True)


def test_path_for_parameter_set():
    class Test(Enum):
        MEMBER = 1

    assert path_for_parameter_set({"foo": "bar"}) == (HERE / "output_foo-bar.h5")
    assert path_for_parameter_set({"enum": Test.MEMBER}) == (HERE / "output_enum-member.h5")


def test_processor_against_reference():
    for parameter_set in PARAMETER_SETS:
        with open(path_for_parameter_set(parameter_set), "rb") as f:
            expected = load_output(f)

        actual = get_output(parameter_set)
        compare_output(expected, actual)


def test_processor_against_reference_calibrated():
    for parameter_set in PARAMETER_SETS_CALIBRATION:
        with open(path_for_parameter_set(parameter_set, calibration=True), "rb") as f:
            expected = load_output(f)

        actual = get_output_precalibrated(parameter_set)
        compare_output(expected, actual)


def test_calibration_against_reference_calibration():
    for parameter_set in PARAMETER_SETS_CALIBRATION:
        expected = np.load(HERE / "calibration.npy")
        actual = get_calibration(parameter_set)
        actual_legacy_format = np.array(
            [
                actual.stationary_clutter_mean,
                actual.stationary_clutter_std,
            ]
        )
        np.testing.assert_array_equal(expected, actual_legacy_format)


@pytest.mark.parametrize("reference_file", HERE.glob("output_*.h5"))
def test_output_references_have_detections(reference_file):
    reference = load_output(reference_file)

    found_peaks = reference.get("found_peaks")
    assert isinstance(found_peaks, np.ndarray)
    assert not np.isnan(found_peaks).all()


@pytest.mark.parametrize("path", ["test.npy", "test.npz"])
def test_load_calibration_from_paths(path):
    DistanceDetectorCalibration.validate_path(path)


@pytest.mark.parametrize(
    "path,expectation",
    [
        ("test.npy", pytest.raises(ValueError)),
        ("test.npz", does_not_raise()),
    ],
)
def test_save_calibration_from_paths(path, expectation):
    with expectation:
        DistanceDetectorCalibration.validate_path(
            path, file_extensions=[("npz", "Numpy data archives (*.npz)")]
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    sp = subparsers.add_parser("save")

    args = parser.parse_args()

    if args.command == "save":
        for parameter_set in PARAMETER_SETS:
            output = get_output(parameter_set)
            save_output(path_for_parameter_set(parameter_set), output)

        for parameter_set in PARAMETER_SETS_CALIBRATION:
            output = get_output_precalibrated(parameter_set)
            save_output(path_for_parameter_set(parameter_set, calibration=True), output)
    else:
        raise RuntimeError



import multiprocessing as mp

import pytest

from acconeer.exptool import a111


def test_mode_to_config_map():
    set(a111.Mode.__members__) == set(a111._configs.MODE_TO_CONFIG_CLASS_MAP.keys())


def test_value_get_set():
    conf = a111.SparseServiceConfig()

    assert conf.downsampling_factor == 1
    conf.downsampling_factor = 2
    assert conf.downsampling_factor == 2
    del conf.downsampling_factor
    assert conf.downsampling_factor == 1


@pytest.mark.parametrize("mode", a111.Mode)
def test_value_dump_load(mode):
    config_cls = a111._configs.MODE_TO_CONFIG_CLASS_MAP[mode]
    config = config_cls()
    assert config.downsampling_factor == 1
    config.downsampling_factor = 2
    dump = config._dumps()

    config = config_cls()
    assert config.downsampling_factor == 1
    config._loads(dump)
    assert config.downsampling_factor == 2

    another_mode = next(other_mode for other_mode in a111.Mode if other_mode != mode)
    another_config_cls = a111._configs.MODE_TO_CONFIG_CLASS_MAP[another_mode]

    config = another_config_cls()
    with pytest.raises(AssertionError):
        config._loads(dump)


def dump_fun(config):
    return config._dumps()


def test_mp():
    conf = a111.SparseServiceConfig()
    conf.downsampling_factor = 2
    [dump] = mp.Pool(1).map(dump_fun, [conf])
    assert dump == conf._dumps()



from . import a111, utils
from .pg_process import PGProccessDiedException, PGProcess
from .structs import configbase


try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0"



import inspect

import pytest

import acconeer.exptool.structs.configbase as cb
from acconeer.exptool.a111 import _configs
from acconeer.exptool.a111._clients.reg import regmap
from acconeer.exptool.a111._modes import Mode


BO = regmap.BYTEORDER


def test_full_names_unique():
    unique_names = set([r.full_name for r in regmap.REGISTERS])
    assert len(regmap.REGISTERS) == len(unique_names)


def test_get_reg_status():
    reg = regmap.get_reg("status")
    assert reg.full_name == "status"
    assert reg == regmap.STATUS_REG
    assert reg.bitset_flags == regmap.STATUS_FLAGS
    assert reg.bitset_masks == regmap.STATUS_MASKS
    assert regmap.get_reg(reg) == reg
    assert regmap.get_reg(reg.addr) == reg


def test_get_reg():
    with pytest.raises(ValueError):
        regmap.get_reg("does-not-exist")

    assert regmap.get_reg("iq_stitch_count").full_name == "iq_stitch_count"
    assert regmap.get_reg("iq_stitch_count", "iq").full_name == "iq_stitch_count"
    assert regmap.get_reg("stitch_count", "iq").full_name == "iq_stitch_count"

    with pytest.raises(ValueError):
        regmap.get_reg("iq_stitch_count", "sparse")

    with pytest.raises(ValueError):
        regmap.get_reg("stitch_count")  # ambiguous

    reg = regmap.get_reg("sp_start")

    with pytest.raises(ValueError):
        regmap.get_reg(reg.addr)  # ambiguous

    assert regmap.get_reg(reg.addr, reg.modes[0]) == reg


def test_config_to_reg_map_completeness():
    all_param_keys = set()

    for mode, config_class in _configs.MODE_TO_CONFIG_CLASS_MAP.items():
        params = {k: v for k, v in inspect.getmembers(config_class) if isinstance(v, cb.Parameter)}
        all_param_keys.update(params.keys())

        expected_config_key_to_reg_map = {}

        for param_name, param in params.items():
            if param.is_dummy:
                continue

            reg_name = regmap.CONFIG_TO_STRIPPED_REG_NAME_MAP[param_name]

            if reg_name is None:
                continue

            reg = regmap.get_reg(reg_name, mode)

            assert reg.category in [regmap.Category.CONFIG, regmap.Category.GENERAL]

            expected_config_key_to_reg_map[param_name] = reg

        assert regmap.get_config_key_to_reg_map(mode) == expected_config_key_to_reg_map

    unknown_keys_in_map = set(regmap.CONFIG_TO_STRIPPED_REG_NAME_MAP.keys()) - all_param_keys
    assert not unknown_keys_in_map


def test_encode_bitset():
    reg = regmap.STATUS_REG
    assert reg.data_type == regmap.DataType.BITSET
    created = regmap.STATUS_FLAGS.CREATED
    activated = regmap.STATUS_FLAGS.ACTIVATED

    truth = int(activated).to_bytes(4, BO)
    assert reg.encode(activated) == truth
    assert reg.encode(int(activated)) == truth
    assert reg.encode("activated") == truth
    assert reg.encode("ACTIVATED") == truth
    assert reg.encode(["activated"]) == truth

    truth = int(0).to_bytes(4, BO)
    assert reg.encode([]) == truth
    assert reg.encode(0) == truth

    truth = int(created | activated).to_bytes(4, BO)
    assert reg.encode(created | activated) == truth
    assert reg.encode(["created", "activated"]) == truth


def test_decode_bitset():
    reg = regmap.STATUS_REG
    created = regmap.STATUS_FLAGS.CREATED
    activated = regmap.STATUS_FLAGS.ACTIVATED

    assert reg.decode(reg.encode(created)) == created
    assert reg.decode(reg.encode(created | activated)) == created | activated


def test_encode_enum():
    reg = regmap.get_reg("mode_selection")
    assert reg.data_type == regmap.DataType.ENUM
    envelope = reg.enum.ENVELOPE

    truth = int(envelope).to_bytes(4, BO)
    assert reg.encode(envelope) == truth
    assert reg.encode(int(envelope)) == truth
    assert reg.encode("envelope") == truth
    assert reg.encode("ENVELOPE") == truth

    # Implicit remapping
    assert reg.encode(Mode.ENVELOPE) == truth

    # Explicit remapping
    reg = regmap.get_reg("repetition_mode")
    truth = int(reg.enum.STREAMING).to_bytes(4, BO)
    assert reg.encode(_configs.BaseServiceConfig.RepetitionMode.SENSOR_DRIVEN) == truth


def test_decode_enum():
    reg = regmap.get_reg("mode_selection")
    envelope = reg.enum.ENVELOPE

    assert reg.decode(reg.encode(envelope)) == envelope


def test_encode_bool():
    reg = regmap.get_reg("tx_disable")
    assert reg.data_type == regmap.DataType.BOOL

    assert reg.encode(False) == int(0).to_bytes(4, BO)
    assert reg.encode(True) == int(1).to_bytes(4, BO)
    assert reg.encode(0) == int(0).to_bytes(4, BO)
    assert reg.encode(1) == int(1).to_bytes(4, BO)
    assert reg.encode(123) == int(1).to_bytes(4, BO)


def test_decode_bool():
    reg = regmap.get_reg("tx_disable")

    assert reg.decode(reg.encode(True)) is True


def test_encode_int():
    pass  # tested in float


def test_decode_int():
    pass


def test_encode_uint():
    reg = regmap.get_reg("downsampling_factor")
    assert reg.data_type == regmap.DataType.UINT32

    assert reg.encode(0) == int(0).to_bytes(4, BO, signed=True)
    assert reg.encode(1234) == int(1234).to_bytes(4, BO, signed=True)

    with pytest.raises(ValueError):
        reg.encode(-123)


def test_decode_uint():
    reg = regmap.get_reg("downsampling_factor")

    assert reg.decode(reg.encode(1234)) == 1234


def test_encode_float():
    reg = regmap.get_reg("range_start")
    assert reg.full_name == "range_start"
    assert reg.float_scale == pytest.approx(1000)
    assert reg.data_type == regmap.DataType.INT32

    assert reg.encode(0) == int(0).to_bytes(4, BO, signed=True)
    assert reg.encode(0.123) == int(123).to_bytes(4, BO, signed=True)
    assert reg.encode(-0.123) == int(-123).to_bytes(4, BO, signed=True)


def test_decode_float():
    reg = regmap.get_reg("range_start")

    assert reg.decode(reg.encode(0.123)) == pytest.approx(0.123)
